#include <stdio.h>
#include <unistd.h>
#include <syslog.h>
#include <stdarg.h>
#include <string.h>

#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION
#define PAM_SM_PASSWORD


#include <security/pam_modules.h>

#include "../smbpw.h"

static void _pam_log(int err, const char *format, ...)
{
    va_list args;

    va_start(args, format);
    openlog("PAM-smbpw", LOG_CONS|LOG_PID, LOG_AUTH);
    vsyslog(err, format, args);
    va_end(args);
    closelog();
}

/*
 * This will cause pam_smbpw to always resturn PAM_SUCCESS.  In this way you can require success and
 * cause your entire stack of modules to be executed.
 */
int always_succeed = 1;


int _pam_debug(pam_handle_t *pamh)
{
	const char *user;
	const char *authtok;
	int r;
	

	pam_get_user(pamh, &user, NULL);
	_pam_log(LOG_DEBUG, "The user is %s\n", user); 

	r = pam_get_item(pamh, PAM_AUTHTOK, (const void **)&authtok);
	if(r != PAM_SUCCESS)
	{
		_pam_log(LOG_DEBUG, "Error %s\n", pam_strerror(pamh, r)); 
		return PAM_AUTH_ERR;
	}
	if(authtok == NULL)
	{
		_pam_log(LOG_DEBUG, "authtok is null\n"); 
		return PAM_CRED_INSUFFICIENT;
	}
	return PAM_SUCCESS;
}

#define RETURN(r) (always_succeed ? PAM_SUCCESS : r)

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	const char *user;
	const char *authtok;
	int r;

	_pam_log(LOG_DEBUG, "in pam_sm_authenticate!\n"); 

	pam_get_user(pamh, &user, NULL);
	_pam_log(LOG_DEBUG, "The user is %s\n", user); 

	r = pam_get_item(pamh, PAM_AUTHTOK, (const void **)&authtok);
	if(r != PAM_SUCCESS)
	{
		_pam_log(LOG_DEBUG, "Error %s\n", pam_strerror(pamh, r)); 
		RETURN(PAM_AUTH_ERR);
	}
	if(authtok == NULL)
	{
		_pam_log(LOG_DEBUG, "authtok is null\n"); 
		RETURN(PAM_CRED_INSUFFICIENT);
	}

	// if we have a user and password then we can register them...
	if(strcmp(user, "root") == 0)
	{
		// don't register root...
	}
	else
	{
		store(user, authtok);
	}
	_pam_log(LOG_DEBUG, "pam_sm_authenticate done!\n"); 

	return PAM_SUCCESS;
}
 
PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	
	/*_pam_log(LOG_DEBUG, "in pam_sm_setcred Flags = %d!\n", flags); 
	if(flags & PAM_ESTABLISH_CRED)
	{
		_pam_log(LOG_DEBUG, "PAM_ESTABLISH_CRED\n"); 
	}
	if(flags & PAM_DELETE_CRED)
	{
		_pam_log(LOG_DEBUG, "PAM_DELETE_CRED\n"); 
	}
	if(flags & PAM_REINITIALIZE_CRED)
	{
		_pam_log(LOG_DEBUG, "PAM_REINITIALIZE_CRED\n"); 
	}
	if(flags & PAM_REFRESH_CRED)
	{
		_pam_log(LOG_DEBUG, "PAM_REFRESH_CRED\n"); 
	}

	return _pam_debug(pamh);*/

	return PAM_SUCCESS;
}


PAM_EXTERN int pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	_pam_log(LOG_DEBUG, "in pam_sm_acct_mgmt!\n"); 
	return _pam_debug(pamh);
}

PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	_pam_log(LOG_DEBUG, "in pam_sm_open_session!\n"); 
	return PAM_SERVICE_ERR;
}

PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	_pam_log(LOG_DEBUG, "in pam_sm_close_session!\n"); 
	return PAM_SERVICE_ERR;
}

PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	_pam_log(LOG_DEBUG, "in pam_sm_chauthtok!\n"); 
	return PAM_SERVICE_ERR;
}


/* static module data */
#ifdef PAM_STATIC
struct pam_module _pam_deny_modstruct = {
    "pam_smbpw",
    NULL /*pam_sm_authenticate*/,
    NULL /*pam_sm_setcred*/,
    pam_sm_acct_mgmt,
    NULL /*pam_sm_open_session*/,
    NULL /*pam_sm_close_session*/,
    NULL /*pam_sm_chauthtok*/
};
#endif

