/*
 * This is a passowrd cache for smb passwords
 *
 */

#include "smbpw.h"

#include <sys/types.h>
#include <sys/wait.h>
 

/*
 *
 * We have to make sure that the path we are going to create
 * the socket in is owned by root and can not be written to
 * by anyone else, otherwise someone might be able to play
 * funny buggers with the socket and subvert the security.
 *
 */
int verify_socket_path(char *name)
{
	int i = 0;
	char x;
	int result = 0;
	struct stat buf;

	name = strdup(name);
	if(!name)
	{
		return -1;
	}

	while(name[i])
	{
		if(name[i] == '/')
		{
			x = name[i + 1];
			name[i + 1] = 0;
			//printf("path = %s\n", name);
			stat(name, &buf);
			//printf("uid = %d gid = %d\n", buf.st_uid, buf.st_gid);
			if(buf.st_uid != 0)
			{
				result = -1;
				fprintf(stderr, "directory %s not owned by root\n", name);
			}
			if(buf.st_mode & S_IWGRP)
			{
				result = -1;
				fprintf(stderr, "group write access to %s\n", name);
			}
			if(buf.st_mode & S_IWOTH)
			{
				result = -1;
				fprintf(stderr, "other write access to %s\n", name);
			}
			name[i + 1] = x;
		}
		i++;
	}

	free(name);
	return result;
}

void start_server()
{
	pid_t cpid;
	int status;

	cpid = fork();
	if(cpid == 0)
	{
		// Detach ourself so PAM does not have a fit
		cpid = fork();

		if(cpid != 0)
		{
			exit(0);
		}
		setsid();

		execl(SERVER_PATH SERVER_NAME, SERVER_NAME, 0);
		exit(0);
	}
	waitpid(cpid, &status, 0);

	return;
}

int open_unix_socket(char *name)
{
	struct stat stat_buf;
	int ss;
	struct sockaddr_un addr;

	if(stat(name, &stat_buf) == 0)
	{
		// if this is a socket then delete it
		if(S_ISSOCK(stat_buf.st_mode))
		{
			if(unlink(name) == -1)
			{
				perror("open_unix_socket:unlink");
			};
		}
	}
	else
	{
		if(errno != ENOENT)
		{
			perror("open_unix_socket:stat");
		}
	}
                
	ss = socket(PF_UNIX, SOCK_STREAM, 0);
	if(ss == -1)
	{
		perror("open_unix_socket:socket");
		return -1;
	}

	addr.sun_family = AF_UNIX;
	strncpy(addr.sun_path, name, UNIX_PATH_MAX);

	if(bind(ss, (struct sockaddr *)&addr, sizeof(addr)) == -1)
	{
		perror("open_unix_socket:bind");
		return -1;
	}

	// listen for connections
	if(listen(ss, 1) == -1)
	{
		perror("open_unix_socket:listen");
		return -1;
	};

	return ss;
}

int unix_connect(char *name)
{
	struct sockaddr_un addr;
	int cs = socket(PF_UNIX, SOCK_STREAM, 0);
	if(cs == -1)
	{
		perror("socket");
		return -1;
	}

	addr.sun_family = AF_UNIX;
	strncpy(addr.sun_path, name, UNIX_PATH_MAX);

	if(connect(cs, (struct sockaddr *)&addr, sizeof(addr)) == -1)
	{
		//perror("connect");
		close(cs);
		return -1;
	}
	return cs;
}

char *readstring(int cs)
{
	char buffer[1024];
	int i = 0;
	int len;
	while(1)
	{
		len = read(cs, buffer + i, sizeof(char));
		if(len > 0)
		{
			if(buffer[i] == 0)
			{
				break;
			}
			i++;
		}
		else
		{
			return 0;
		}
	}

	return strdup(buffer);
}

int store(char *username, char *password)
{
	int cs;
	char command;
	ssize_t len;
	char result = 0xff;
	int attempts = 3;
	int delay = 1;
	char *domain_user;


	cs = unix_connect(SOCKET_NAME);
	if(cs < 0)
	{
		start_server();
		while(attempts > 0)
		{
			attempts--;
			sleep(delay);
			delay = delay * 2;

			cs = unix_connect(SOCKET_NAME);
			if(cs >= 0)
			{
				break;
			}
		}
	}

	if(cs < 0)
	{
		// could not connect to server
		return -1;
	}

	// see if there is a domain to strip off
	domain_user = strchr(username, '+');
	if(domain_user)
	{
		username = domain_user + 1;
	}
	// store
	command = 0x00;
	write(cs, &command, sizeof(command));
	write(cs, username, strlen(username) + 1);
	write(cs, password, strlen(password) + 1);
	// read the result
	len = read(cs, &result, sizeof(result));
	if(len > 0)
	{
		//printf("Got result 0x%02x\n", result);
	}
	else
	{
		//printf("read len = %d\n", len);
		if(len == -1)
		{
			perror("read");
		}
	}
	close(cs);

	return 0;
}

char *retrieve(char *username)
{
	return real_retrieve(username, 0x01);
}

char *query(char *username)
{
	return real_retrieve(username, 0x03);
}

char *real_retrieve(char *username, char command)
{
	int cs;
	ssize_t len;
	char result = 0xff;
	char *password = 0;

	cs = unix_connect(SOCKET_NAME);

	if(cs < 0)
	{
		// could not connect, try to start the server...
		start_server();
		cs = unix_connect(SOCKET_NAME);
	}

	write(cs, &command, sizeof(command));
	write(cs, username, strlen(username) + 1);
	// read the result
	len = read(cs, &result, sizeof(result));
	if(len > 0)
	{
		//printf("Got result 0x%02x\n", result);

		if(result == 0x00)
		{
			password = readstring(cs);
			//printf("The password is %s\n", password);
		}
	}
	else
	{
		// error?
		//printf("read len = %d\n", len);
		if(len == -1)
		{
			perror("read");
		}
	}
	close(cs);
	
	return password;
}

int shutdown_server(void)
{
	int cs;
	char command;
	char result = 0x00;

	cs = unix_connect(SOCKET_NAME);

	// store
	command = 0x02;
	write(cs, &command, sizeof(command));
	// no result
	close(cs);

	return result;
}


